unit FMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, XPMan, StdCtrls, ComCtrls, IniFiles,
  ImagingComponents, ImagingUtility, ExtDlgs, Buttons, Spin;

type
  TMainF = class(TForm)
    XPManifest1: TXPManifest;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel5: TPanel;
    Image1: TImage;
    OpenDialog: TOpenPictureDialog;
    BtnRun: TBitBtn;
    BtnExit: TBitBtn;
    PageControl: TPageControl;
    TabSettings: TTabSheet;
    Panel4: TPanel;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    LabelSize: TLabel;
    LabelMemory: TLabel;
    EditHeightmap: TEdit;
    Button3: TButton;
    ComboSize: TComboBox;
    EditTexture: TEdit;
    Button4: TButton;
    ComboPreset: TComboBox;
    ComboBlocks: TComboBox;
    GroupBox2: TGroupBox;
    CheckFullscreen: TCheckBox;
    CheckDXT: TCheckBox;
    CheckAnisoTex: TCheckBox;
    CheckFog: TCheckBox;
    CheckDetail: TCheckBox;
    TabOutput: TTabSheet;
    BtnBench: TBitBtn;
    MemoOut: TMemo;
    SpinScale: TSpinEdit;
    Label8: TLabel;
    SpinX: TSpinEdit;
    SpinY: TSpinEdit;
    Label9: TLabel;
    Label10: TLabel;
    CheckFilter: TCheckBox;
    CheckAnisoError: TCheckBox;
    procedure ComboBlocksChange(Sender: TObject);
    procedure ComboSizeChange(Sender: TObject);
    procedure ComboPresetChange(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure BtnRunClick(Sender: TObject);
    procedure BtnExitClick(Sender: TObject);
    procedure BtnBenchClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    procedure ShowError(const Msg: string);
    procedure ShowDialog(Edit: TEdit);
    procedure SetEditText(Edit: TEdit; const Text: string);
    procedure RecalculateTerrainSize;
    function GetBlocks: LongInt;
    function GetTerrainSize: LongInt;
    procedure LineOut(const Line: string);
    procedure RunApp(Benchmark: Boolean);
  public

  end;

const
  SAppName = 'glSOAR';
  SAppExe = 'glSOAR.exe';
  SIniFile = 'Launcher.ini';
  STrue = 'true';
  SFalse = 'false';
  SDataDir = 'Data' + PathDelim;
  SHeightVolcanoes = SDataDir + 'Volcanoes.png';
  STexVolcanoes = SDataDir + 'VolcanoesTX.dds';
  SHeightEverest = SDataDir + 'Everest.png';
  STexEverest = SDataDir + 'EverestTX.dds';
  SHeightPuget = SDataDir + 'Puget.png';
  STexPuget = SDataDir + 'PugetTX.dds';
  SHeightDesert = SDataDir + 'Desert.png';
  STexDesert = SDataDir + 'DesertTX.dds';
  VertexSize = 20;

var
  MainF: TMainF;

implementation

{$R *.dfm}

procedure TMainF.BtnBenchClick(Sender: TObject);
begin
  RunApp(True);
end;

procedure TMainF.BtnExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainF.LineOut(const Line: string);
begin
  MemoOut.Lines.Append(StringReplace(Line, #13, '', []));
end;

procedure TMainF.BtnRunClick(Sender: TObject);
begin
  RunApp(False);
end;

procedure TMainF.FormClose(Sender: TObject; var Action: TCloseAction);
var
  Ini: TIniFile;
begin
  try
    Ini := TIniFile.Create(GetAppDir + PathDelim + SIniFile);

    Ini.WriteString('Terrain', 'Heightmap', EditHeightmap.Text);
    Ini.WriteString('Terrain', 'GroundTex', EditTexture.Text);
    Ini.WriteInteger('Terrain', 'Size', GetTerrainSize);
    Ini.WriteInteger('Terrain', 'Blocks', GetBlocks);
    Ini.WriteInteger('Terrain', 'Scale', SpinScale.Value);
    Ini.WriteInteger('Terrain', 'StartX', SpinX.Value);
    Ini.WriteInteger('Terrain', 'StartY', SpinY.Value);
    Ini.WriteBool('Terrain', 'Filter8bit', CheckFilter.Checked);
    Ini.WriteBool('Terrain', 'AnisoMetric', CheckAnisoError.Checked);

    Ini.WriteBool('Graphic', 'Fullscreen', CheckFullscreen.Checked);
    Ini.WriteBool('Graphic', 'DXT', CheckDXT.Checked);
    Ini.WriteBool('Graphic', 'Detail', CheckDetail.Checked);
    Ini.WriteBool('Graphic', 'AnisoTex', CheckAnisoTex.Checked);
    Ini.WriteBool('Graphic', 'Fog', CheckFog.Checked);

    Ini.Free;
  except
  end;

  Action := caFree;
end;

procedure TMainF.FormCreate(Sender: TObject);
var
  Ini: TIniFile;
  Size, Blocks: LongInt;

  procedure SetFileToEdit(Edit: TEdit; const FileName, Default: string);
  begin
    if (Pos(':', FileName) > 0) and FileExists(FileName) then
      SetEditText(Edit, FileName)
    else
    if FileExists(GetAppDir + PathDelim + FileName) then
      SetEditText(Edit, GetAppDir + PathDelim + FileName)
    else
      SetEditText(Edit, Default);
  end;

  procedure FindInCombo(Combo: TComboBox; Num: LongInt);
  var
    I: LongInt;
    S: string;
  begin
    S := IntToStr(Num);

    for I := 0 to Combo.Items.Count - 1 do
    begin
      if Pos(S, Combo.Items[I]) > 0 then
      begin
        Combo.ItemIndex := I;
        Exit;
      end;
    end;
  end;

begin
  try
    Ini := TIniFile.Create(GetAppDir + PathDelim + SIniFile);

    EditHeightmap.Text := Ini.ReadString('Terrain', 'Heightmap', '');
    EditTexture.Text := Ini.ReadString('Terrain', 'GroundTex', '');

    SetFileToEdit(EditHeightmap, EditHeightmap.Text, GetAppDir + PathDelim + SHeightVolcanoes);
    SetFileToEdit(EditTexture, EditTexture.Text, GetAppDir + PathDelim + STexVolcanoes);

    Size := Ini.ReadInteger('Terrain', 'Size', 1025);
    Blocks := Ini.ReadInteger('Terrain', 'Blocks', 1);

    FindInCombo(ComboSize, Size);
    FindInCombo(ComboBlocks, Blocks);

    RecalculateTerrainSize;

    SpinScale.Value := Ini.ReadInteger('Terrain', 'Scale', 25);
    SpinX.Value := Ini.ReadInteger('Terrain', 'StartX', 0);
    SpinY.Value := Ini.ReadInteger('Terrain', 'StartY', 0);
    CheckFilter.Checked := Ini.ReadBool('Terrain', 'Filter8bit', True);
    CheckAnisoError.Checked := Ini.ReadBool('Terrain', 'AnisoMetric', True);

    CheckFullscreen.Checked := Ini.ReadBool('Graphic', 'Fullscreen', True);
    CheckDXT.Checked := Ini.ReadBool('Graphic', 'DXT', True);
    CheckDetail.Checked := Ini.ReadBool('Graphic', 'Detail', True);
    CheckAnisoTex.Checked := Ini.ReadBool('Graphic', 'AnisoTex', True);
    CheckFog.Checked := Ini.ReadBool('Graphic', 'Fog', False);

    Ini.Free;
  except
    ComboPreset.ItemIndex := 0;
    ComboPresetChange(Self);
    RecalculateTerrainSize;
  end;
end;

procedure TMainF.ShowError(const Msg: string);
begin
  MessageBox(Handle, PChar(Msg), 'Error', MB_ICONERROR or MB_OK);
end;

procedure TMainF.Button3Click(Sender: TObject);
begin
  ShowDialog(EditHeightmap);
end;

procedure TMainF.Button4Click(Sender: TObject);
begin
  ShowDialog(EditTexture);
end;

procedure TMainF.ShowDialog(Edit: TEdit);
begin
  OpenDialog.Filter := Graphics.GraphicFilter(TImagingGraphic);
  OpenDialog.FileName := Edit.Text;
  if OpenDialog.Execute then
  begin
    SetEditText(Edit, OpenDialog.FileName);
  end;
end;

procedure TMainF.ComboPresetChange(Sender: TObject);
var
  HeightMap, TexMap: string;
begin
  // Settings for all presets
  ComboSize.ItemIndex := 3; // 1km x 1km
  SpinX.Value := 0;
  SpinY.Value := 0;
  case ComboPreset.ItemIndex of
    0:
      begin
        // Volcanoes
        HeightMap := GetAppDir + PathDelim + SHeightVolcanoes;
        TexMap := GetAppDir + PathDelim + STexVolcanoes;
        SpinScale.Value := 25;
        SpinX.Value := 4;
        SpinY.Value := 16;
      end;
    1:
      begin
        // Everest
        HeightMap := GetAppDir + PathDelim + SHeightEverest;
        TexMap := GetAppDir + PathDelim + STexEverest;
        SpinScale.Value := 25;
        SpinX.Value := 3;
        SpinY.Value := 11;
      end;
    2:
      begin
        // Puget Sound
        HeightMap := GetAppDir + PathDelim + SHeightPuget;
        TexMap := GetAppDir + PathDelim + STexPuget;
        SpinScale.Value := 5;
        SpinX.Value := 14;
        SpinY.Value := 6;
      end;
    3:
      begin
        // Dune
        HeightMap := GetAppDir + PathDelim + SHeightDesert;
        TexMap := GetAppDir + PathDelim + STexDesert;
        SpinScale.Value := 15;
      end;
  end;
  SetEditText(EditHeightmap, HeightMap);
  SetEditText(EditTexture, TexMap);
end;

procedure TMainF.SetEditText(Edit: TEdit; const Text: string);
begin
  Edit.Text := Text;
  Edit.SelStart := Length(Edit.Text);
end;

procedure TMainF.RecalculateTerrainSize;
var
  Blocks, BlockSize, TotalSize, NumVertices, NumIndices: LongInt;
begin
  Blocks := GetBlocks;
  BlockSize := GetTerrainSize;
  TotalSize := BlockSize * Blocks;
  LabelSize.Caption := Format('%.0n x %.0n', [TotalSize * 1.0, TotalSize * 1.0]);
  NumVertices := BlockSize * BlockSize;
  NumIndices :=  BlockSize * BlockSize * 4;
  LabelMemory.Caption := Format('%.1n MiB', [(NumVertices * VertexSize) / (1024 * 1024)]);
end;

procedure TMainF.RunApp(Benchmark: Boolean);
var
  Exe, CmdLine, CurrentLine: string;
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  Security: TSecurityAttributes;
  BytesRead, Code: LongWord;
  Buffer: array[0..255] of Char;
  HOutputReadTmp, HOutputRead, HOutputWrite, HInputWriteTmp, HInputRead,
    HInputWrite, HErrorWrite: THandle;
begin
  PageControl.ActivePage := TabOutput;
  BtnRun.Enabled := False;
  BtnBench.Enabled := False;
  Exe := GetAppDir + PathDelim + SAppExe;
  CmdLine := Exe;

  if not FileExists(EditHeightmap.Text) then
  begin
    ShowError('Heightmap "' + EditHeightmap.Text + '" does not exist.');
    Exit;
  end;

  if not Benchmark then
  begin
    CmdLine := CmdLine + ' -map="'        + EditHeightmap.Text + '"';
    CmdLine := CmdLine + ' -texture"='    + EditTexture.Text + '"';
    CmdLine := CmdLine + ' -size='        + IntToStr(GetTerrainSize);
    CmdLine := CmdLine + ' -blocks='      + IntToStr(GetBlocks);
    CmdLine := CmdLine + ' -scale='       + IntToStr(SpinScale.Value);
    CmdLine := CmdLine + ' -startx='      + IntToStr(SpinX.Value);
    CmdLine := CmdLine + ' -starty='      + IntToStr(SpinY.Value);
    CmdLine := CmdLine + ' -anisometric=' + Iff(CheckDXT.Checked, STrue, SFalse);
    CmdLine := CmdLine + ' -filter8bit='  + Iff(CheckFilter.Checked, STrue, SFalse);

    CmdLine := CmdLine + ' -fullscreen='  + Iff(CheckFullscreen.Checked, STrue, SFalse);
    CmdLine := CmdLine + ' -anisotex='    + Iff(CheckAnisoTex.Checked, STrue, SFalse);
    CmdLine := CmdLine + ' -dxt='         + Iff(CheckDXT.Checked, STrue, SFalse);
    CmdLine := CmdLine + ' -fog='         + Iff(CheckFog.Checked, STrue, SFalse);
    CmdLine := CmdLine + ' -detail='      + Iff(CheckDetail.Checked, STrue, SFalse);
  end
  else
  begin
    CmdLine := CmdLine + ' -benchmark=true';
  end;

  MemoOut.Lines.Clear;
  LineOut('Command line:');
  LineOut(StringReplace(CmdLine, GetAppDir + PathDelim, '', [rfReplaceAll, rfIgnoreCase]));
  LineOut('');
  LineOut('Program output:');

  Security.nLength := SizeOf(TSecurityAttributes);
  Security.lpSecurityDescriptor := nil;
  Security.bInheritHandle := True;

  CreatePipe(HOutputReadTmp, HOutputWrite, @Security, 0);
  DuplicateHandle(GetCurrentProcess(), HOutputWrite, GetCurrentProcess(),
    @HErrorWrite, 0, True, DUPLICATE_SAME_ACCESS);
  CreatePipe(HInputRead, HInputWriteTmp, @Security, 0);

  DuplicateHandle(GetCurrentProcess(), HOutputReadTmp, GetCurrentProcess(),
    @HOutputRead, 0, False, DUPLICATE_SAME_ACCESS);
  DuplicateHandle(GetCurrentProcess(), HInputWriteTmp, GetCurrentProcess(),
    @HInputWrite, 0, False, DUPLICATE_SAME_ACCESS);
  CloseHandle(HOutputReadTmp);
  CloseHandle(HInputWriteTmp);

  FillChar(StartupInfo, SizeOf(StartupInfo), 0);
  with StartupInfo do
  begin
    cb := SizeOf(StartupInfo);
    lpTitle := PChar(SAppName);
    dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW or STARTF_FORCEONFEEDBACK;
    hStdInput := HInputRead;
    hStdOutput := HOutputWrite;
    hStdError := HErrorWrite;
    wShowWindow := SW_SHOWNORMAL;
    // when ShowWindow is 0 console is not opened, but glut window is not opened too
  end;

  if not CreateProcess(nil, PChar(CmdLine), @Security, @Security, True,
    NORMAL_PRIORITY_CLASS, nil, PChar(GetAppDir), StartupInfo, ProcessInfo) then
  begin
    ShowError('Cannot run ' + SAppExe);
    Exit;
  end;

  CloseHandle(HOutputWrite);
  CloseHandle(HInputRead);
  CloseHandle(HErrorWrite);

  // this halts app even - must change to asynchronous ReadFile
  repeat
    if (not ReadFile(HOutputRead, Buffer, 1, BytesRead, nil)) or (BytesRead = 0) then
    begin
      if GetLastError = ERROR_BROKEN_PIPE then
        Break;
    end;
    Buffer[BytesRead] := #0;
    CurrentLine := CurrentLine + PChar(@Buffer[0]);
    if CurrentLine[Length(CurrentLine)] = #10 then
    begin
      Delete(CurrentLine, Length(CurrentLine) - 1, 2);
      LineOut(CurrentLine);
      CurrentLine := '';
    end;
    Application.ProcessMessages;
  until False;

  GetExitCodeProcess(ProcessInfo.hProcess, Code);

  LineOut('');
  LineOut('Program terminated with exit code ' + IntToStr(Code));

  CloseHandle(HOutputRead);
  CloseHandle(HInputWrite);
  CloseHandle(ProcessInfo.hProcess);
  CloseHandle(ProcessInfo.hThread);

  BtnRun.Enabled := True;
  BtnBench.Enabled := True;
end;

procedure TMainF.ComboSizeChange(Sender: TObject);
begin
  RecalculateTerrainSize;
end;

procedure TMainF.ComboBlocksChange(Sender: TObject);
begin
  RecalculateTerrainSize;
end;

function TMainF.GetBlocks: LongInt;
begin
  Result := StrToInt(StringReplace(
    Copy(ComboBlocks.Items[ComboBlocks.ItemIndex], 1, 2), ' ', '', [rfReplaceAll]));
end;

function TMainF.GetTerrainSize: LongInt;
begin
  Result := Pow2Int(ComboSize.ItemIndex + 7) + 1
end;

end.
